/*
 * Environment.cpp
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#include "Environment.h"

Environment::Environment(IsoGaussian* covfunc, Mat &Field, int N, int m)
{
	this->covfunc = covfunc; this->Field = Field;
	this->N = N; this->nR = m;
	this->K = this->covfunc->Eval(Field);
	this->mean = Vec(N, 0.0);
	this->Y = MultiGaussRan(this->mean, this->K, this->nR);
}

double Environment::Observe(int idx, int version) // request an observation at location index idx for the corresponding version of the environment's realization
{
	return Y[version][idx];
}

double Environment::RMS(Vec &ID, int v, vector <IsoGaussian*> &kernels, Vec &b)
{
	double res = 0.0, predict = 0.0;

	// initialize the corresponding set of kernels

	vector <GP*> models; models.clear();

	for (int t = 0; t < (int) kernels.size(); t++)
	{
		GP* model = new GP(kernels[t]);
		models.push_back(model);
	}

	// prepare the training data

	Mat X; X.clear(); Vec Z; Z.clear();
	bool* marker = new bool[N];
	for (int i = 0; i < N; i++) marker[i] = false;

	for (int i = 0; i < (int) ID.size(); i++)
	{
		marker[(int) ID[i]] = true; // all locations correspond to the index set have been selected
		X.push_back(Field[(int) ID[i]]); // construct the set of observed locations
		Z.push_back(Y[v][(int) ID[i]]); // construct the corresponding set of observed measurements
	}

	// train each GP model

	for (int t = 0; t < (int) kernels.size(); t++)
		models[t]->train(X, Z);

	// evaluate performance

	for (int s = 0; s < N; s++)
		if (!marker[s]) // consider each unobserved location
		{
			predict = 0.0;

			for (int t = 0; t < (int) kernels.size(); t++) // for each kernel
			{
				pair <double, double> ans = models[t]->pointwise_prediction(Field[s]); // individual prediction
				predict += (b[t] * ans.first); // update the average prediction
			}

			res += pow(predict - Y[v][s], 2.0); // update the sum of square error
		}

	return sqrt(res / (N - (int)ID.size())); // root mean square error
}


