/*
 * Factory.cpp
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#include "Factory.h"

Factory::Factory(Mat &Field, int N, vector <IsoGaussian*> &covfuncs, int nL)
{
	this->Field = Field; this->N = N;
	this->covfuncs = covfuncs; this->nL = nL;
	this->b = Vec(this->nL, 1.0 / this->nL);


	this->offset = INFTY;
	for (int t = 0; t < this->nL; t++)
		this->offset = min(this->offset, this->covfuncs[t]->Noise());
	this->offset = -min(0.0, ENT(pow(this->offset, 2.0)));

	this->m = 0;

	this->X.clear(); this->K.clear(); this->L.clear();
	this->marker = new bool[this->N];
	for (int i = 0; i < this->N; i++) this->marker[i] = false;

	this->models = vector <GP*> (this->nL);

	for (int t = 0; t < this->nL; t++)
	{
		this->K.push_back(Mat(0));
		this->L.push_back(Mat(0));
		this->models[t] = new GP(this->covfuncs[t]);
 	}

	this->pX = Vec(0); this->pY = Vec(0);
}

void Factory::Initialize(Vec &X, int k)
{
	this->X = X; this->m = (int) this->X.size();

	for (int i = 0; i < this->N; i++) this->marker[i] = false;
	for (int i = 0; i < this->m; i++) this->marker[(int) this->X[i]] = true;

	this->K[k].clear(); this->L[k].clear();
	if (this->m == 0) return;

	this->K[k] = Mat(this->m, Vec(this->m, 0.0));
	for (int i = 0; i < this->m; i++)
		for (int j = 0; j < this->m; j++) this->K[k][i][j] = this->covfuncs[k]->Eval(this->Field[this->X[i]], this->Field[this->X[j]]);

	this->L[k] = Cholesky(this->K[k]);
}

double Factory::Entropy(int idx, int k)
{
	double var = this->covfuncs[k]->Var(this->Field[idx]);

	if (this->m > 0)
	{
		Vec kk(this->m, 0.0);
		for (int j = 0; j < this->m; j++) kk[j] = this->covfuncs[k]->Eval(this->Field[idx], this->Field[X[j]]);
		Vec v = ForwardSub(this->L[k], kk);
		var = var - Dot(v, v);
	}

	return ENT(var);
}

pair <double, int> Factory::NextLocation(int k)
{
	pair <double, int> res = make_pair(-INFTY, -1);

	for (int i = 0; i < this->N; i++)
		if (!this->marker[i])
		{
			double entropy = Entropy(i, k);
			if (entropy > res.first)
			{
				res.first = entropy;
				res.second = i;
			}
		}

	return res;
}

pair <double, Vec> Factory::GreedyDesign(Vec &X, int n, int k)
{
	Initialize(X, k);

	n = min(n, this->N - this->m);

	pair <double, Vec> res = make_pair(0.0, Vec(n, -1));

	for (int i = 0; i < n; i++)
	{
		pair <double, int> next = NextLocation(k);
		res.first += next.first;
		res.second[i] = next.second;

		this->marker[next.second] = true; this->X.push_back(next.second);

		for (int j = 0; j < this->m; j++) K[k][j].push_back(this->covfuncs[k]->Eval(Field[next.second], Field[this->X[j]]));
		this->K[k].push_back(Vec(this->m + 1, 0.0));
		for (int j = 0; j < this->m; j++) K[k][this->m][j] = K[k][j][this->m];
		this->K[k][this->m][this->m] = this->covfuncs[k]->Var(Field[next.second]);

		this->m = this->m + 1;
		this->L[k] = Cholesky(this->K[k]);
	}

	return res;
}

double Factory::Entropy(int idx)
{
	double res = 0.0, var = 0.0;

	for (int t = 0; t < this->nL; t++)
	{
		var = this->covfuncs[t]->Var(this->Field[idx]);

		if (this->m > 0)
		{
			Vec kk(this->m, 0.0);
			for (int j = 0; j < this->m; j++) kk[j] = this->covfuncs[t]->Eval(this->Field[idx], this->Field[X[j]]);
			Vec v = ForwardSub(this->L[t], kk);
			var -= Dot(v, v);
		}

		res = res + this->b[t] * ENT(var);
	}

	return res;
}

pair <double, int> Factory::NextLocation()
{
	pair <double, int> res = make_pair(-INFTY, -1);

	for (int i = 0; i < N; i++)
		if (!this->marker[i])
		{
			double entropy = Entropy(i);
			if (entropy > res.first)
			{
				res.first = entropy;
				res.second = i;
			}
		}

	return res;
}

void Factory::Initialize(Vec &X)
{
	this->X = X; this->m = (int) this->X.size();

	for (int i = 0; i < this->N; i++) this->marker[i] = false;
	for (int i = 0; i < this->m; i++) this->marker[(int) X[i]] = true;

	this->K.clear(); this->L.clear();
	for (int t = 0; t < this->nL; t++)
	{
		this->K.push_back(Mat(0));
		this->L.push_back(Mat(0));
	}

	if (this->m == 0) return;

	for (int t = 0; t < this->nL; t++)
	{
		this->K[t] = Mat(this->m, Vec(this->m, 0.0));
		for (int i = 0; i < this->m; i++)
			for (int j = 0; j < this->m; j++) this->K[t][i][j] = this->covfuncs[t]->Eval(this->Field[X[i]], this->Field[X[j]]);
		this->L[t] = Cholesky(this->K[t]);
	}
}

pair <double, Vec> Factory::GreedyDesign(Vec &X, int n, Vec &b)
{
	Initialize(X);

	n = min(n, this->N - this->m); this->b = b;

	pair <double, Vec> res = make_pair(0.0, Vec(n, -1));

	for (int i = 0; i < n; i++)
	{
		pair <double, int> next = NextLocation();
		res.first += next.first;
		res.second[i] = next.second;

		this->marker[next.second] = true; this->X.push_back(next.second);

		for (int t = 0; t < this->nL; t++)
		{
			for (int j = 0; j < this->m; j++) K[t][j].push_back(this->covfuncs[t]->Eval(this->Field[next.second], this->Field[this->X[j]]));
			this->K[t].push_back(Vec(this->m + 1, 0.0));
			for (int j = 0; j < this->m; j++) this->K[t][this->m][j] = this->K[t][j][this->m]; this->K[t][this->m][this->m] = this->covfuncs[t]->Var(this->Field[next.second]);
			this->L[t] = Cholesky(this->K[t]);
		}

		this->m = this->m + 1;
	}

	return res;
}

pair <double, double> Factory::InitialBounds(Vec &X, Vec &b, int n)
{
	this->m = (int) X.size(); this->b = b;
	n = min(n, this->N - this->m);

	if (n == 0) return make_pair(0.0, 0.0);

	pair <double, Vec> lb = GreedyDesign(X, n, b);
	vector < pair <double, Vec> > ubs(this->nL);

	double LB = lb.first, UB = 0.0;

	for (int t = 0; t < this->nL; t++)
		if (b[t] > EPS) UB -= b[t] * log(b[t]);

	for (int t = 0; t < this->nL; t++)
	{
		ubs[t] = GreedyDesign(X, n, t);
		UB += b[t] * ((E * ubs[t].first + n * this->offset) / (E - 1));
	}

	return make_pair(LB, UB);
}

double Factory::SubmodularConstant()
{
	return this->offset;
}

vector < vector < pair <double, double> > > Factory::PosteriorGP(Vec &X, Vec &Y)
{
	Mat XX(0); this->m = (int) X.size();

	for (int i = 0; i < this->N; i++) this->marker[i] = false;
	for (int i = 0; i < this->m; i++)
	{
		this->marker[(int)X[i]] = true;
		XX.push_back(this->Field[X[i]]);
	}

	for (int t = 0; t < this->nL; t++) this->models[t]->train(XX, Y);

	vector < vector < pair <double, double> > > res(0);

	for (int i = 0; i < this->N; i++)
		if (!this->marker[i])
		{
			vector < pair <double, double> > post (this->nL);
			for (int t = 0; t < this->nL; t++) post[t] = models[t]->pointwise_prediction(this->Field[i]);
			res.push_back(post);
		}

	return res;
}

vector < pair <int, Vec> > Factory::Sample(Vec &X, vector < vector < pair <double, double> > > &pGP, Vec &b, int nSample)
{
	for (int i = 0; i < this->N; i++) this->marker[i] = false; this->m = (int) X.size();
	for (int i = 0; i < this->m; i++) this->marker[(int) X[i]] = true;

	int index = 0, k = 0;

	vector < pair <int, Vec> > samples(0);

	for (int i = 0; i < this->N; i++)
		if (!this->marker[i])
		{
			pair <int, Vec> tuple(i, Vec(nSample, 0.0));

			for (int j = 0; j < nSample; j++)
			{
				k = DiscreteSample(b);
				tuple.second[j] = GRAND(pGP[index][k].first, sqrt(pGP[index][k].second));
			}

			samples.push_back(tuple); index = index + 1;
		}

	return samples;
}

void Factory::GetPriorData(Vec &X, Vec &Y)
{
	X = this->pX; Y = this->pY;
}

void Factory::UpdatePriorData(int x, double y)
{
	this->pX.push_back(x); this->pY.push_back(y);
}

vector < pair <double, double> > Factory::PosteriorGP(Vec &X, Vec &Y, int x)
{
	Mat XX(0); this->m = (int) X.size();

	for (int i = 0; i < this->m; i++) XX.push_back(this->Field[X[i]]);

	for (int t = 0; t < this->nL; t++) this->models[t]->train(XX, Y);

	vector < pair <double, double> > res(this->nL);
	for (int t = 0; t < this->nL; t++) res[t] = models[t]->pointwise_prediction(this->Field[x]);

	return res;
}

