/*
 * IELearner.h
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#ifndef IELEARNER_H_
#define IELEARNER_H_

#include "IsoGaussian.h"

class IELearner
{
	private:

	Mat Field; // the entire field
	int N, nL; // the size of the field and the number of candidate kernels
	bool *marker; // marker of whether a location has been selected previously
	Vec b; // the current distribution over candidate kernels

	Vec X, Y; // X: the set of indices of selected locations; Y: the corresponding measurements
	int m; // the number of selected locations so far

	vector <IsoGaussian*> covfunc; // the set of candidate kernels
	vector <Mat> K, L; // corresponding covariance matrices and their Cholesky decompositions

	double offset, MLB;

	pair <double, int> GreedyChoice(int kernel);
	double Entropy(int s, int kernel);

	pair <double, int> GreedyChoice();
	double Entropy(int s);

	double Likelihood(int selected, double obs, int t);

	public:

	IELearner(Mat &Field, int N, vector <IsoGaussian*> &covfunc , int nL, Vec &b);

	void Refresh(Vec &b);

	pair <double, Vec> AprioriDesign(int n, int kernel); // compute the a priori design for the next n observations, as well as their absorbed entropy, given the current data and a specific kernel
	pair <double, Vec> AprioriDesign(int n); // compute the a priori design for the next n observations, as well as their absorbed entropy, given the current data

	void Update(int selected, double obs); // update the covariance matrices, the belief over possible kernels as well as the Cholesky decompositions
	void LightUpdate(int selected, double obs);

	Vec Move(int n, double stop);

	Vec GetBelief();
};

#endif /* IELEARNER_H_ */
