/*
 * IsoGaussian.cpp
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#include "IsoGaussian.h"

IsoGaussian::IsoGaussian(double sig, double band, double sn)
{
    this->sig = sig; this->band = band; this->sn = sn;
}

double IsoGaussian::Eval(double x, double y)
{
	if (fabs(x - y) <= EPS) return pow(this->sig, 2.0) * exp(-pow(x - y, 2.0) / (2.0 * pow(this->band, 2.0))) + pow(this->sn, 2.0); else
		return pow(this->sig, 2.0) * exp(-pow(x - y, 2.0) / (2.0 * pow(this->band, 2.0)));
}

Mat IsoGaussian::Eval(Vec &X)
{
    int N = (int) X.size();
    Mat S(N, Vec(N, 0.0));

    for (int i = 0; i < N; i++)
        for (int j = 0; j < N; j++) S[i][j] = Eval(X[i], X[j]);

    return S;
}

double IsoGaussian::Eval(Vec &x, Vec &y)
{
    int N = (int) x.size();

    double covar = 0.0;

    for (int i = 0; i < N; i++) covar -= pow(x[i] - y[i], 2.0);

    if (fabs(covar) <= EPS) return pow(this->sig, 2.0) * exp(covar / (2.0 * pow(this->band, 2.0))) + pow(this->sn, 2.0); else
    	return pow(this->sig, 2.0) * exp(covar / (2.0 * pow(this->band, 2.0)));
}

Mat IsoGaussian::Eval(Mat & X)
{
    int N = (int) X.size();

    Mat S(N, Vec(N, 0.0));

    for (int i = 0; i < N; i++)
        for (int j = 0; j < N; j++) S[i][j] = Eval(X[i], X[j]);

    return S;
}

Vec IsoGaussian::Eval(double x, Vec &Y)
{
    int N = (int) Y.size();
    Vec X(N, 0.0);

    for (int i = 0; i < N; i++) X[i] = Eval(x, Y[i]);
    return X;
}

Vec IsoGaussian::Eval(Vec &x, Mat &Y)
{
    int N = (int) Y.size();
    Vec X(N, 0.0);

    for (int i = 0; i < N; i++) X[i] = Eval(x, Y[i]);
    return X;
}

double IsoGaussian::Var(double x)
{
    return pow(this->sig, 2.0) + pow(this->sn, 2.0);
}

double IsoGaussian::Var(Vec &x)
{
    return pow(this->sig, 2.0) + pow(this->sn, 2.0);
}

double IsoGaussian::Noise()
{
	return sn;
}

double IsoGaussian::Signal()
{
	return sig;
}

double IsoGaussian::Bandwidth()
{
	return band;
}


