/*
 * RTDP.h
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#ifndef RTDP_H_
#define RTDP_H_

#include "GP.h"
#include "Node.h"
#include "Factory.h"

struct NodeComparator
{
	bool operator() (Node* a, Node* b)
	{
		pair <double, double> aBounds = a->GetBounds(), bBounds = b->GetBounds();
		double aGap = fabs(aBounds.second - aBounds.first), bGap = fabs(bBounds.second - bBounds.first);

		if (aBounds.first > bBounds.first) return false; else
			if ((fabs(aBounds.first - bBounds.first) <= EPS) && (aGap > bGap)) return false; // lower-bound first, then comes the gap

		/*if (aBounds.second > bBounds.second) return false; else
			if ((fabs(aBounds.second - bBounds.second) <= EPS) && (aGap < bGap)) return false; // upper-bound first, then comes the gap*/

		/*if (aGap > bGap) return false; else
			if ((fabs(aGap - bGap) <= EPS) && (a->Budget() >= b->Budget())) return false;*/
		return true;
	}
};

class RTDP
{
	private:

	Mat Field; // the entire field
	int N, nL; // the size of the field and the number of kernels

	vector <IsoGaussian*> covfuncs; // the set of kernels
	Vec b; // current belief over possible kernels

	Node *root;

	priority_queue <Node*, vector <Node*>, NodeComparator> Q;

	Node* Explore(Node* node, double threshold);

	Factory* factory;

	public:

	RTDP(Mat &Field, int N, vector <IsoGaussian*> &covfuncs, int nL, Vec &b);

	Node* Plan(int n, double threshold, int maxpath);

	void UpdateData(int x, double y);

	Vec GetBelief();
};

#endif /* RTDP_H_ */
