#include "Environment.h"
#include "Factory.h"
#include "IsoGaussian.h"
#include "IELearner.h"
#include "GP.h"
#include "Node.h"
#include "RTDP.h"
#include "ITELearner.h"

void factory_test() // okay
{
	srand(51987);

	double band = 1.0, sig = 5.0, sn = 0.25;
	int nL = 5;
	Vec b(nL, 1.0/nL);

	vector <IsoGaussian*> covfunc(nL);
	for (int t = 0; t < nL; t++)
		covfunc[t] = new IsoGaussian(sig, band * (t + 1), sn);

	int N = 100, k = 2, n = 40;

	Mat Field(N, Vec(1, 0.0));
	for (int i = 0; i < N; i++) Field[i][0] = i * 1.0;

	Factory* factory = new Factory(Field, N, covfunc, nL);
	Vec X(0);

	cout << "Performance Bounds : " << n * ENT(sn * sn) << " " << n * ENT(sn * sn + sig * sig) << endl;

	pair <double, Vec> res = factory->GreedyDesign(X, n, b);
	cout << "Absorbed Entropy : " << res.first << endl; cout << "Sampling Sites : ";
	for (int i = 0; i < (int) res.second.size(); i++) cout << res.second[i] << " "; cout << endl;

	res = factory->GreedyDesign(X, n, b);
	cout << "Absorbed Entropy : " << res.first << endl; cout << "Sampling Sites : ";
	for (int i = 0; i < (int) res.second.size(); i++) cout << res.second[i] << " "; cout << endl;

	res = factory->GreedyDesign(X, n, k);
	cout << "Absorbed Entropy : " << res.first << endl; cout << "Sampling Sites : ";
	for (int i = 0; i < (int) res.second.size(); i++) cout << res.second[i] << " "; cout << endl;

	IELearner* learner = new IELearner(Field, N, covfunc, nL, b);
	res = learner->AprioriDesign(n);
	cout << "Absorbed Entropy : " << res.first << endl; cout << "Sampling Sites : ";
	for (int i = 0; i < (int) res.second.size(); i++) cout << res.second[i] << " "; cout << endl;
}

void environment_test() // okay
{
	srand(51987);

	double band = 1.0, sig = 5.0, sn = 0.25;
	int nL = 5;
	Vec b(nL, 1.0/nL);

	vector <IsoGaussian*> covfunc(nL);
	for (int t = 0; t < nL; t++)
		covfunc[t] = new IsoGaussian(sig, band * (t + 1), sn);

	int N = 100, nR = 100, kernel = 1;

	Mat Field(N, Vec(1, 0.0));
	for (int i = 0; i < N; i++) Field[i][0] = i * 1.0;

	Environment* env = new Environment(covfunc[kernel], Field, N, nR);

	Vec Y(N, 0.0);

	double zMin = INFTY, zMax = -INFTY;

	for (int i = 1; i < N; i++)
	{
		for (int v = 0; v < nR; v++)
		{
			Y[i] += env->Observe(i, v);
			zMin = min(zMin, env->Observe(i, v));
			zMax = max(zMax, env->Observe(i, v));
		}
		Y[i] = Y[i] / nR;
	}

	for (int i = 1; i < N; i++) cout << Y[i] << " "; cout << endl;

	cout << zMin << " " << zMax << endl;
}

void ielearner_test() // okay
{
	srand(51987);

	double band = 1.0, sig = 10.0, sn = 0.25;
	int nL = 5;
	Vec b(nL, 1.0/nL);

	vector <IsoGaussian*> covfunc(nL);
	covfunc[0] = new IsoGaussian(sig, 2 * band, sn);
	for (int t = 1; t < nL; t++)
		covfunc[t] = new IsoGaussian(sig, band * 4 * (t + 1), sn);

	int N = 100, k = 0, n = 50, nR = 20, m = n;

	Mat Field(N, Vec(1, 0.0));
	for (int i = 0; i < N; i++) Field[i][0] = i * 1.0;

	Environment* env = new Environment(covfunc[k], Field, N, nR);

	IELearner* learner = new IELearner(Field, N, covfunc, nL, b);

	pair <double, Vec> res = learner->AprioriDesign(n);

	double set_rms = 0.0, policy_rms = 0.0, stop = 0.35;
	Vec loc, pb, next;

	cout << "Evaluating The A Priori Design ... " << endl;

	for (int v = 0; v < nR; v++)
	{
		cout << "Iteration " << v << " ." << endl;
		set_rms += env->RMS(res.second, v, covfunc, b);
	}

	cout << "Evaluating The Adaptive Policy ..." << endl;

	for (int v = 0; v < nR; v++)
	{
		cout << "Iteration " << v << " ." << endl;
		loc.clear();
		learner->Refresh(b); n = m;

		while (n > 0)
		{
			next = learner->Move(n, stop);
			if ((int) next.size() == 1) learner->Update(next[0], env->Observe(next[0], v));
			for (int i = 0; i < (int) next.size(); i++) loc.push_back(next[i]);
			n = n - next.size();
		}

		pb = learner->GetBelief();
		policy_rms += env->RMS(loc, v, covfunc, pb);
	}

	cout << "Greedy A Priori Design RMS = " << set_rms / nR << endl;
	cout << "Greedy Policy RMS = " << policy_rms / nR << endl;

	double zMin = INFTY, zMax = -INFTY;

	for (int v = 0; v < nR; v++)
		for (int i = 0; i < N; i++)
		{
			zMin = min(zMin, env->Observe(i, v));
			zMax = max(zMax, env->Observe(i, v));
		}

	cout << "Measurement Range : " << zMin << " " << zMax << endl;
}

void entropy_approx_test() // okay
{
	srand(51982237);
	double sigma = 2.0, mean = 0.0, var = pow(sigma, 2.0);
	int n = 8;

	Vec samples = grand(mean, sigma, n);

	for (int i = 0; i < n; i++) cout << samples[i] << " "; cout << endl;

	double ent = ENT(var), emp = 0.0;

	for (int i = 0; i < n; i++)
	{
		cout << -LN(samples[i], mean, var) << endl;
		emp -= LN(samples[i], mean, var);
	}

	emp /= n;

	cout << "Entropy = " << ent << endl;
	cout << "Empirical Entropy = " << emp << endl;

	cout << -LN(1000, mean, var) << endl;
}

void node_test() // okay
{
	srand(51987);

	double band = 1.0, sig = 5.0, sn = 0.25;
	int nL = 5;
	Vec b(nL, 1.0/nL);

	vector <IsoGaussian*> covfunc(nL);
	for (int t = 0; t < nL; t++)
		covfunc[t] = new IsoGaussian(sig, band * (t + 1), sn);

	int N = 100, n = 5;

	Mat Field(N, Vec(1, 0.0));
	for (int i = 0; i < N; i++) Field[i][0] = i * 1.0;

	Factory* factory = new Factory(Field, N, covfunc, nL);

	Node* node = new Node(n, 0, 0, NULL, factory, b);
	pair <double, double> bounds = node->GetBounds();

	cout << bounds.first << " " << bounds.second << endl;

	node->Explore();

	node->UpdateBounds();

	bounds = node->GetBounds();

	cout << bounds.first << " " << bounds.second << endl;
}

void RTDP_test() // okay
{
	srand(51987);

	double band = 1.0, sig = 10.0, sn = 0.25;
	int nL = 7;
	Vec b(nL, 1.0/nL), bands(nL);

	bands[0] = 1.0 * band;
	for (int t = 1; t < nL; t++) bands[t] = band * 3 * (t + 1);

	vector <IsoGaussian*> covfunc(nL);
	for (int t = 0; t < nL; t++)
		covfunc[t] = new IsoGaussian(sig, bands[t], sn);

	int N = 100, n = 3, nPath = 1000;

	double threshold = 0.1;

	Mat Field(N, Vec(1, 0.0));
	for (int i = 0; i < N; i++) Field[i][0] = i * 1.0;

	srand(51987);

	RTDP* planner = new RTDP(Field, N, covfunc, nL, b);
	Node* res = planner->Plan(n, threshold, nPath);

	pair <double, double> bounds = res->GetBounds();

	cout << endl;
	cout << "The Final Bound : " << bounds.first << " " << bounds.second << endl;
}

void performance_test()
{
	srand(51987);

	double band = 1.0, sig = 10.0, sn = 0.25;
	int nL = 7;
	Vec b(nL, 1.0/nL), bands(nL);

    bands[0] = 1.0 * band;
    for (int t = 1; t < nL; t++) bands[t] = band * 3 * (t + 1);

	vector <IsoGaussian*> covfunc(nL);
	for (int t = 0; t < nL; t++)
		covfunc[t] = new IsoGaussian(sig, bands[t], sn);

	int N = 100, k = 0, n = 10, nR = 20, m = n;

	Mat Field(N, Vec(1, 0.0));
	for (int i = 0; i < N; i++) Field[i][0] = i * 1.0;

	Environment* env = new Environment(covfunc[k], Field, N, nR);

	IELearner* learner = new IELearner(Field, N, covfunc, nL, b);

	pair <double, Vec> res = learner->AprioriDesign(n);

	double set_rms = 0.0, policy_rms = 0.0, stop = 0.05, zMin = INFTY, zMax = -INFTY;
	Vec loc, pb, next;

	for (int v = 0; v < nR; v++)
	for (int i = 0; i < N; i++)
	{
		zMin = min(zMin, env->Observe(i, v));
		zMax = max(zMax, env->Observe(i, v));
	}

	cout << "Evaluating The A Priori Greedy Design ... " << endl;

	for (int v = 0; v < nR; v++)
	{
		cout << "Iteration " << v << " ." << endl;

		learner->Refresh(b);
		for (int i = 0; i < (int) res.second.size(); i++)
		{
			cout << res.second[i] << " ";
			learner->Update(res.second[i], env->Observe(res.second[i], v));
		}

		cout << endl;

		pb = learner->GetBelief();

		for (int t = 0; t < nL; t++) cout << pb[t] << " "; cout << endl;

		set_rms += env->RMS(res.second, v, covfunc, pb);
	}

	cout << "Evaluating The Adaptive Greedy Policy ..." << endl;

	for (int v = 0; v < nR; v++)
	{
		cout << "Iteration " << v << " ." << endl;
		loc.clear();
		learner->Refresh(b); n = m;

		while (n > 0)
		{
			next = learner->Move(n, stop);
			if ((int) next.size() == 1) learner->Update(next[0], env->Observe(next[0], v));
			for (int i = 0; i < (int) next.size(); i++) loc.push_back(next[i]);
			n = n - (int) next.size();
		}

		if ((int) next.size() > 1)
			for (int i = 0; i < (int) next.size(); i++) learner->Update(next[i], env->Observe(next[i], v));

		for (int i = 0; i < (int) loc.size(); i++)
			cout << loc[i] << " "; cout << endl;

		pb = learner->GetBelief();

		for (int t = 0; t < nL; t++) cout << pb[t] << " "; cout << endl;

		policy_rms += env->RMS(loc, v, covfunc, pb);
	}

	int step = 2, nPath = 100, x;
	double threshold = 2 * BIAS, rtdp = 0.0;

	srand(51987);

	cout << "Evaluating The Adaptive RTDP Policy ..." << endl;

	Factory* fact = new Factory(Field, N, covfunc, nL);

	for (int v = 0; v < nR; v++)
	{
		cout << "Iteration " << v << " ." << endl;
		loc.clear(); n = m;

		RTDP* planner = new RTDP(Field, N, covfunc, nL, b);

		while (n > 0)
		{
			Node* node = planner->Plan(step, threshold, nPath);
			x = node->NextLocation(); loc.push_back(x);

			cout << v << " " << x << " " << n - 1 << endl;

			planner->UpdateData(x, env->Observe(x, v));
			n = n - 1;
			delete node;

			pb = planner->GetBelief();

            double maxp = 0.0; int selected = 0;

            for (int t = 0; t < nL; t++)
            	if (maxp < pb[t])
            	{
            		maxp = pb[t];
            		selected = t;
            	}

            if (step >= 3 && maxp >= 0.9)
            {
            	next = fact->GreedyDesign(loc, n, selected).second;
            	for (int i = 0; i < (int) next.size(); i++)
            	{
            		cout << v << " " << next[i] << " " << n - 1 << endl;
            		loc.push_back(next[i]); n = n - 1;
            		planner->UpdateData(next[i], env->Observe(next[i], v));
            	}
            }
		}

		pb = planner->GetBelief();

		for (int t = 0; t < nL; t++) cout << pb[t] << " "; cout << endl;

		rtdp += env->RMS(loc, v, covfunc, pb);
	}

    cout << "Experiment Settings : " << endl;
    cout << "--" << endl;
    cout << "The Total Number of Sampling Sites : " << N << endl;
    cout << "Measurement Ranges : " << "[" << zMin << ", " << zMax << "]" << endl << endl;
    cout << "*******************" << endl << endl;

    cout << "Correlation Settings : " << endl;
    cout << "--" << endl;
    cout << "Signal and Noise : " << sig << " and " << sn << endl;
    cout << "The Set of Correlation Bandwidths : (" << nL << ") :" << endl;
    for (int i = 0; i < nL; i++) cout << bands[i] << " "; cout << endl;
    cout << "True Bandwidth : " << bands[k] << endl;
    cout << "Prior Belief : " << endl;
    for (int i = 0; i < nL; i++) cout << b[i] << " "; cout << endl << endl;
    cout << "*******************" << endl << endl;

    cout << "Algorithm Settings : " << endl;
    cout << "--" << endl;
    cout << "Sensing Budget : " << m << endl;
    cout << "RTDP Planning Horizon : " << step << endl;
    cout << "RTDP Exploration Path : " << nPath << endl << endl;
    cout << "*******************" << endl << endl;

    cout << "Performance Results (averaged over " << nR << " realizations) : " << endl;
    cout << "--" << endl;
	cout << "Greedy A Priori Design RMS = " << set_rms / nR << endl;
	cout << "Greedy Adaptive Policy RMS = " << policy_rms / nR << endl;
	cout << "RTDP Adaptive Policy RMS = " << rtdp / nR  << endl;

}

void test_bandwidth()
{
	double ell[7], alpha = 0.2, a = 0.1;
	ell[0] = 1; ell[1] = 6; ell[2] = 9; ell[3] = 12; ell[4] = 15; ell[5] = 18; ell[6] = 21;

	for (int i = 0; i < 6; i++)
	{
		double delta = 2 * ell[i] * ell[i + 1] * sqrt(log(ell[i + 1] / ell[i]) / (ell[i + 1] * ell[i + 1] - ell[i] * ell[i]));
		double rho = exp(-(delta * delta) / (ell[i + 1] * ell[i + 1])) - exp(-(delta * delta) / (ell[i] * ell[i]));

		double r1 = 1.0 / (rho * rho), r2 = log(1.0 / alpha) * log(1.0 / alpha);

		cout << ell[i] << " " << delta << " " << rho << " " << r1 * r2 << " " << ceil(a * r1 * r2) + 1 << endl;
	}
}

void ite_test()
{
	srand(51987);

	double band = 1.0, sig = 10.0, sn = 0.25;
	int nL = 7;
	Vec b(nL, 1.0/nL), bands(nL);

	bands[0] = 1.0 * band;
	for (int t = 1; t < nL; t++) bands[t] = band * 3 * (t + 1);

	vector <IsoGaussian*> covfunc(nL);
	for (int t = 0; t < nL; t++)
		covfunc[t] = new IsoGaussian(sig, bands[t], sn);

	int N = 100, k = 0, n = 15, nR = 20, m = n;

	Mat Field(N, Vec(1, 0.0));
	for (int i = 0; i < N; i++) Field[i][0] = i * 1.0;

	Environment* env = new Environment(covfunc[k], Field, N, nR);

	IELearner* learner = new IELearner(Field, N, covfunc, nL, b);

	cout << "Evaluating The Adaptive ITE Policy ..." << endl;

	double policy_rms = 0.0; Vec loc;

	for (int v = 0; v < nR; v++)
	{
		cout << "Iteration " << v << " ." << endl;

		ITELearner* ite = new ITELearner(Field, N, covfunc, nL, b);

		loc.clear();
		learner->Refresh(b); n = m;

		while (n > 0)
		{
			Vec next = ite->Move(n);
			Vec obs(0);

			for (int i = 0; i < (int) next.size(); i++)
			{
				obs.push_back(env->Observe(next[i], v));
				loc.push_back(next[i]);
			}

			n = n - (int) next.size();

			if (n > 0) ite->Update(next, obs);
		}

		for (int i = 0; i < (int) loc.size(); i++)
		{
			cout << loc[i] << " ";
			learner->Update(loc[i], env->Observe(loc[i], v));
		}

		cout << endl;

		Vec pb = learner->GetBelief();

		for (int t = 0; t < nL; t++) cout << pb[t] << " "; cout << endl;

		policy_rms += env->RMS(loc, v, covfunc, pb);

	}

	cout << endl;

	cout << "ITE Policy RMS : " << policy_rms / nR << endl;

}

int main()
{
	//environment_test();

	//factory_test();

	//ielearner_test();

	//entropy_approx_test();

	//node_test();

	//RTDP_test();

	//performance_test();

	//traffic_test("aim_slot_1");

	//ite_test();

	//cout << exp(9.0/42) << endl;

	return 0;
}
