/*
 * utility.h
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#ifndef UTILITY_H_
#define UTILITY_H_

#include <iostream>
#include <cstdio>
#include <cmath>
#include <ctime>
#include <cstring>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include <vector>
#include <set>
#include <queue>
#include <deque>
#include <list>
#include <map>
#include <iomanip>
#include <stack>
#include <bitset>
#include <algorithm>
#include <functional>
#include <numeric>

using namespace std;

#define Vec vector <double>
#define Mat vector < Vec >
#define IRAND(a, b) (rand() % ((b) - (a) + 1) + (a)) // randomly generate an integer from [a, b]
#define DRAND(a, b) (((b) - (a)) * ((double) rand() / (RAND_MAX)) + (a)) // randomly generate a real number from [a, b]
#define GRAND(mu, sigma) ((mu) + (sigma) * grand()) // randomly generate a real number from N(mu, sigma^2)
#define EPS 0.00000000000000001
#define PI 3.141592653589793238462
#define E 2.71828182845904509080
#define INFTY 1000000000.0
#define N(m, mean, var) ((1 / sqrt(2 * PI * (var))) * exp(-0.5 * (pow((m) - (mean), 2.0) / (var)))) // the pdf function for N(mean, var = sigma^2)
#define ENT(var) (0.5 * log(2.0 * PI * E * (var))) // the entropy of N(., var)
#define LN(m, mean, var) (-0.5 * log(2 * PI * (var)) - 0.5 * (pow((m) - (mean), 2.0)) / (var) ) // the log pdf function for N(mean, var = sigma^2)

Mat Cholesky(Mat &A); // Cholesky decompositon
Vec ForwardSub(Mat &A, Vec &b); // solving Ax = b for a lower triangular matrix A
Vec BackwardSub(Mat &A, Vec &b); // solving Ax = b for a upper triangular matrix B
double Dot(Vec &a, Vec &b); // dot product between two vectors
Vec RanVec(int N, double a, double b, bool integral); // generate a random vector of size N whose (real or integer) components fall between a and b
Mat RanMat(int M, int N, double a, double b, bool integral); // generate a random matrix of size M * N whose (real or integer) components fall between a and b
Mat Mult(Mat &A, Mat &B); // multiply two matrices of matching size
Mat Trans(Mat &A); // transpose a matrix

double grand(); // sample from N(0, 1^2)
Vec grand(double mean, double sigma, int N); // generate n samples from N(mean, sigma^2)
Vec MultiGaussRan(Vec &mean, Mat &cov); // sample a multivariate sample from N(mean, cov)
Mat MultiGaussRan(Vec &mean, Mat &cov, int N); // generate n multivariate samples from N(mean, cov)
int DiscreteSample(Vec &b); // sample from a discrete distribution


#endif /* UTILITY_H_ */
